from django.contrib.contenttypes.models import ContentType
from django.db.models.signals import post_save, pre_save

try:
    from django.contrib.contenttypes.fields import GenericRelation
except ImportError:
    from django.contrib.contenttypes.generic import GenericRelation

from river.models import State, Proceeding
from river.models.proceeding_track import ProceedingTrack
from river.models.managers.wofkflow_object import WorkflowObjectManager
from river.services.object import ObjectService
from river.services.transition import TransitionService

__author__ = 'ahmetdal'

from django.db import models


class StateField(models.ForeignKey):
    def __init__(self, state_model=State, object_manager=WorkflowObjectManager, *args, **kwargs):
        kwargs['null'] = True
        kwargs['blank'] = True
        self.object_manager = object_manager
        kwargs['to'] = '%s.%s' % (state_model._meta.app_label, state_model._meta.object_name)
        super(StateField, self).__init__(*args, **kwargs)

    def contribute_to_class(self, cls, name, virtual_only=False):
        def is_workflow_completed(workflow_object):
            return ObjectService.is_workflow_completed(workflow_object, name)

        def proceed(self, user, *args, **kwargs):
            TransitionService.proceed(self, name, user, *args, **kwargs)

        @property
        def on_initial_state(self):
            from river.services.state import StateService

            return StateService.get_initial_state(ContentType.objects.get_for_model(self), name) == getattr(self, name)

        @property
        def on_final_state(self):
            from river.services.state import StateService

            return getattr(self, name) in StateService.get_final_states(ContentType.objects.get_for_model(self), name)

        def get_initial_state(self):
            from river.services.state import StateService

            return StateService.get_initial_state(ContentType.objects.get_for_model(self), name)

        def get_available_proceedings(self, *args, **kwargs):
            from river.services.proceeding import ProceedingService

            return ProceedingService.get_available_proceedings(self, name, [getattr(self, name)], *args, **kwargs)

        @property
        def initial_proceedings(self):
            from river.services.proceeding import ProceedingService

            return getattr(self, name) in ProceedingService.get_initial_proceedings(ContentType.objects.get_for_model(self), name)

        @property
        def final_proceedings(self):
            from river.services.proceeding import ProceedingService

            return getattr(self, name) in ProceedingService.get_final_proceedings(ContentType.objects.get_for_model(self), name)

        @property
        def next_proceedings(self):
            from river.services.proceeding import ProceedingService

            return getattr(self, name) in ProceedingService.get_next_proceedings(ContentType.objects.get_for_model(self), name)

        @property
        def proceeding_track(self):
            try:
                return ProceedingTrack.objects.filter(proceeding__in=self.proceedings.all()).latest('date_created')
            except ProceedingTrack.DoesNotExist:
                return None

        self.model = cls

        self.__add_to_class(cls, "proceedings", GenericRelation('%s.%s' % (Proceeding._meta.app_label, Proceeding._meta.object_name)))
        self.__add_to_class(cls, "proceeding_track", proceeding_track)

        self.__add_to_class(cls, "objects", self.object_manager(name))
        self.__add_to_class(cls, "is_workflow_completed", is_workflow_completed)
        self.__add_to_class(cls, "proceed", proceed)

        self.__add_to_class(cls, "on_initial_state", on_initial_state)
        self.__add_to_class(cls, "on_final_state", on_final_state)

        self.__add_to_class(cls, "get_initial_state", get_initial_state)
        self.__add_to_class(cls, "get_available_proceedings", get_available_proceedings)

        self.__add_to_class(cls, "initial_proceedings", initial_proceedings)
        self.__add_to_class(cls, "final_proceedings", final_proceedings)
        self.__add_to_class(cls, "next_proceedings", next_proceedings)

        super(StateField, self).contribute_to_class(cls, name, virtual_only=virtual_only)

        post_save.connect(_post_save, self.model, False, dispatch_uid='%s_%s_riverstatefield_post' % (self.model, self.name))
        # self.model.__metaclass__ = WorkflowObjectMetaclass

    def get_state(self, instance):
        return instance.__dict__[self.attname]

    def set_state(self, instance, state):
        instance.__dict__[self.attname] = state.pk

    def __add_to_class(self, cls, key, value):
        if not hasattr(cls, key):
            cls.add_to_class(key, value)


def _post_save(sender, instance, created, *args, **kwargs):  # signal, sender, instance):
    """
    Desc:  Generate TransitionProceedings according to ProceedingMeta of the content type at the beginning.
    :param kwargs:
    :return:
    """
    from river.services.state import StateService

    for f in instance._meta.fields:
        if isinstance(f, StateField):
            if created:
                ObjectService.register_object(instance, f.name)
            if not getattr(instance, f.name):
                init_state = StateService.get_initial_state(ContentType.objects.get_for_model(instance), f.name)
                setattr(instance, f.name, init_state)
                instance.save()



from river.models import State
from river.services.object import ObjectService
from river.services.state import StateService
from river.tests.services.proceeding_service_based_test import ProceedingServiceBasedTest

__author__ = 'ahmetdal'


class test__StateService(ProceedingServiceBasedTest):
    def test_get_available_states(self):
        ObjectService.register_object(self.objects[0], self.field)
        ObjectService.register_object(self.objects[1], self.field)

        available_states = StateService.get_available_states(self.objects[0], self.field, self.user2, include_user=False)
        self.assertEqual(1, available_states.count())
        self.assertEqual(State.objects.get(label='s2'), available_states[0])
        available_states = StateService.get_available_states(self.objects[0], self.field, self.user2, include_user=True)
        self.assertEqual(0, available_states.count())
        available_states = StateService.get_available_states(self.objects[0], self.field, self.user2)
        self.assertEqual(0, available_states.count())

        available_states = StateService.get_available_states(self.objects[0], self.field, self.user3, include_user=False)
        self.assertEqual(1, available_states.count())
        self.assertEqual(State.objects.get(label='s2'), available_states[0])
        available_states = StateService.get_available_states(self.objects[0], self.field, self.user3, include_user=True)
        self.assertEqual(0, available_states.count())
        available_states = StateService.get_available_states(self.objects[0], self.field, self.user3)
        self.assertEqual(0, available_states.count())

        available_states = StateService.get_available_states(self.objects[0], self.field, self.user4, include_user=False)
        self.assertEqual(1, available_states.count())
        self.assertEqual(State.objects.get(label='s2'), available_states[0])
        available_states = StateService.get_available_states(self.objects[0], self.field, self.user4, include_user=True)
        self.assertEqual(0, available_states.count())
        available_states = StateService.get_available_states(self.objects[0], self.field, self.user4)
        self.assertEqual(0, available_states.count())

        available_states = StateService.get_available_states(self.objects[0], self.field, self.user1, include_user=False)
        self.assertEqual(1, available_states.count())
        self.assertEqual(State.objects.get(label='s2'), available_states[0])
        available_states = StateService.get_available_states(self.objects[0], self.field, self.user1, include_user=True)
        self.assertEqual(1, available_states.count())
        self.assertEqual(State.objects.get(label='s2'), available_states[0])

        available_states = StateService.get_available_states(self.objects[0], self.field, self.user1)
        self.assertEqual(1, available_states.count())
        self.assertEqual(State.objects.get(label='s2'), available_states[0])

    def test_get_initial_state(self):
        self.assertEqual(State.objects.get(label='s1'), StateService.get_initial_state(self.content_type, self.field))

    def test_get_final_states(self):
        self.assertListEqual(list(State.objects.filter(label__in=['s4.1', 's4.2', 's5.1', 's5.2'])), list(StateService.get_final_states(self.content_type, self.field)))



from django.db import models

from django.db.models.signals import m2m_changed, post_save

from django.utils.translation import ugettext_lazy as _
from mptt.fields import TreeForeignKey, TreeManyToManyField
from mptt.managers import TreeManager
from mptt.models import MPTTModel

from river.models.base_model import BaseModel
from river.models.transition import Transition
from river.services.config import RiverConfig

__author__ = 'ahmetdal'


class ProceedingMetaManager(models.Manager):
    def get_by_natural_key(self, content_type, field, transition, order):
        return self.get(content_type=content_type, field=field, transition=transition, order=order)


class ProceedingMeta(BaseModel):
    class Meta:
        app_label = 'river'
        verbose_name = _("Proceeding Meta")
        verbose_name_plural = _("Proceeding Metas")
        unique_together = [('content_type', 'field', 'transition', 'order')]

    objects = ProceedingMetaManager()

    content_type = models.ForeignKey(RiverConfig.CONTENT_TYPE_CLASS, verbose_name=_('Content Type'))
    field = models.CharField(verbose_name=_('Field'), max_length=200)

    transition = models.ForeignKey(Transition, verbose_name=_('Transition'))
    permissions = models.ManyToManyField(RiverConfig.PERMISSION_CLASS, verbose_name=_('Permissions'))
    groups = models.ManyToManyField(RiverConfig.GROUP_CLASS, verbose_name=_('Groups'))
    order = models.IntegerField(default=0, verbose_name=_('Order'))
    action_text = models.TextField(_("Action Text"), max_length=200, null=True, blank=True)

    parents = models.ManyToManyField('self', verbose_name='parents', related_name='children', symmetrical=False, db_index=True)

    def natural_key(self):
        return self.content_type, self.field, self.transition, self.order

    def __unicode__(self):
        return 'Transition:%s, Permissions:%s, Order:%s' % (self.transition, ','.join(self.permissions.values_list('name', flat=True)), self.order)


def post_group_change(sender, instance, *args, **kwargs):
    from river.services.proceeding import ProceedingService
    from river.models.proceeding import PENDING

    for proceeding_pending in instance.proceedings.filter(status=PENDING):
        ProceedingService.override_groups(proceeding_pending, instance.groups.all())


def post_permissions_change(sender, instance, *args, **kwargs):
    from river.models.proceeding import PENDING
    from river.services.proceeding import ProceedingService

    for proceeding_pending in instance.proceedings.filter(status=PENDING):
        ProceedingService.override_permissions(proceeding_pending, instance.permissions.all())


def post_save_model(sender, instance, *args, **kwargs):
    from river.services.proceeding_meta import ProceedingMetaService
    ProceedingMetaService.build_tree(instance)


m2m_changed.connect(post_group_change, sender=ProceedingMeta.groups.through)
m2m_changed.connect(post_permissions_change, sender=ProceedingMeta.permissions.through)

post_save.connect(post_save_model, sender=ProceedingMeta)



from unittest import skip
from river.handlers.transition import TransitionHandler, PostTransitionHandler
from river.models import State, Proceeding
from river.services.object import ObjectService
from river.services.transition import TransitionService
from river.tests.services.proceeding_service_based_test import ProceedingServiceBasedTest

__author__ = 'ahmetdal'


class test_TransitionHandler(ProceedingServiceBasedTest):
    @skip("workflow object is now required to register")
    def test_register_for_all(self):
        self.test_args = None
        self.test_kwargs = None

        def test_handler(*args, **kwargs):
            self.test_args = args
            self.test_kwargs = kwargs

        ObjectService.register_object(self.objects[0], self.field)
        ObjectService.register_object(self.objects[1], self.field)

        PostTransitionHandler.register(test_handler)

        self.assertIsNone(self.test_args)
        self.assertIsNone(self.test_kwargs)

        TransitionService.proceed(self.objects[0], self.field, self.user1)

        self.assertEqual((), self.test_args)
        self.assertDictEqual(
            {
                'field': 'my_field',
                'object': self.objects[0],
                'proceeding': Proceeding.objects.filter(meta__transition__source_state=State.objects.get(label='s1'), meta__transition__destination_state=State.objects.get(label='s2'))[0],
                'source_state': State.objects.get(label='s1'),
                'destination_state': State.objects.get(label='s2')
            }, self.test_kwargs)


        # Proceeded but no transition
        TransitionService.proceed(self.objects[0], self.field, self.user2)

        self.assertEqual((), self.test_args)
        self.assertDictEqual(
            {
                'field': 'my_field',
                'object': self.objects[0],
                'proceeding': Proceeding.objects.filter(meta__transition__source_state=State.objects.get(label='s1'), meta__transition__destination_state=State.objects.get(label='s2'))[0],
                'source_state': State.objects.get(label='s1'),
                'destination_state': State.objects.get(label='s2')
            }, self.test_kwargs)

        TransitionService.proceed(self.objects[0], self.field, self.user3)

        self.assertEqual((), self.test_args)
        self.assertDictEqual(
            {
                'field': 'my_field',
                'object': self.objects[0],
                'proceeding': Proceeding.objects.filter(meta__transition__source_state=State.objects.get(label='s2'), meta__transition__destination_state=State.objects.get(label='s3'))[2],
                'source_state': State.objects.get(label='s2'),
                'destination_state': State.objects.get(label='s3')
            }, self.test_kwargs)

    def test_register_for_an_object(self):
        self.test_args = None
        self.test_kwargs = None

        def test_handler(*args, **kwargs):
            self.test_args = args
            self.test_kwargs = kwargs

        ObjectService.register_object(self.objects[0], self.field)
        ObjectService.register_object(self.objects[1], self.field)

        PostTransitionHandler.register(test_handler, self.objects[1], 'my_field')

        self.assertIsNone(self.test_args)
        self.assertIsNone(self.test_kwargs)

        TransitionService.proceed(self.objects[0], self.field, self.user1)

        self.assertIsNone(self.test_args)
        self.assertIsNone(self.test_kwargs)

        TransitionService.proceed(self.objects[0], self.field, self.user2)

        self.assertIsNone(self.test_args)
        self.assertIsNone(self.test_kwargs)

        PostTransitionHandler.register(test_handler, self.objects[0], field='my_field')
        TransitionService.proceed(self.objects[0], self.field, self.user3)

        self.assertEqual((self.objects[0], 'my_field'), self.test_args)
        self.assertDictEqual(
            {
                'proceeding': Proceeding.objects.filter(meta__transition__source_state=State.objects.get(label='s2'), meta__transition__destination_state=State.objects.get(label='s3'))[2],
                'source_state': State.objects.get(label='s2'),
                'destination_state': State.objects.get(label='s3')
            }, self.test_kwargs)

    def test_register_for_a_transition(self):
        self.test_args = None
        self.test_kwargs = None

        def test_handler(*args, **kwargs):
            self.test_args = args
            self.test_kwargs = kwargs

        ObjectService.register_object(self.objects[0], self.field)
        ObjectService.register_object(self.objects[1], self.field)

        PostTransitionHandler.register(test_handler, self.objects[0], 'my_field', source_state=State.objects.get(label='s2'), destination_state=State.objects.get(label='s3'))

        self.assertIsNone(self.test_args)
        self.assertIsNone(self.test_kwargs)

        TransitionService.proceed(self.objects[0], self.field, self.user1)

        self.assertIsNone(self.test_args)
        self.assertIsNone(self.test_kwargs)

        TransitionService.proceed(self.objects[0], self.field, self.user2)

        self.assertIsNone(self.test_args)
        self.assertIsNone(self.test_kwargs)

        TransitionService.proceed(self.objects[0], self.field, self.user3)

        self.assertEqual((self.objects[0], 'my_field'), self.test_args)
        self.assertDictEqual(
            {
                'proceeding': Proceeding.objects.filter(meta__transition__source_state=State.objects.get(label='s2'), meta__transition__destination_state=State.objects.get(label='s3'))[2],
                'source_state': State.objects.get(label='s2'),
                'destination_state': State.objects.get(label='s3')
            }, self.test_kwargs)



from uuid import uuid4
from django.db import models
from django.utils.translation import ugettext_lazy as _
from river.models import BaseModel

__author__ = 'ahmetdal'


class StateManager(models.Manager):
    def get_by_natural_key(self, slug):
        return self.get(slug=slug)


class State(BaseModel):
    class Meta:
        app_label = 'river'
        verbose_name = _("State")
        verbose_name_plural = _("States")

    objects = StateManager()

    slug = models.SlugField(unique=True, null=True, blank=True)
    label = models.CharField(max_length=50)
    description = models.CharField(_("Description"), max_length=200, null=True, blank=True)

    def __unicode__(self):
        return self.label

    def natural_key(self):
        return (self.slug,)

    def details(self):
        detail = super(State, self).details()
        detail.update(
            {
                'label': self.label,
                'description': self.description
            }
        )
        return detail



from river.handlers.backends.loader import load_handler_backend
from river.handlers.transition import PostTransitionHandler
from river.models import Handler
from river.services.config import RiverConfig
from river.services.object import ObjectService
from river.tests.services.proceeding_service_based_test import ProceedingServiceBasedTest

__author__ = 'ahmetdal'


def test_handler(*args, **kwargs):
    pass


class test_DatabaseHandlerBackend(ProceedingServiceBasedTest):
    def setUp(self):
        super(test_DatabaseHandlerBackend, self).setUp()

        RiverConfig.HANDLER_BACKEND_CLASS = 'river.handlers.backends.database.DatabaseHandlerBackend'
        self.handler_backend = load_handler_backend()
        self.handler_backend.handlers = {}

        ObjectService.register_object(self.objects[0], self.field)

    def test_register(self):
        self.assertEqual(0, Handler.objects.count())
        self.assertFalse('%s.%s_object%sfield%s' % (PostTransitionHandler.__module__, PostTransitionHandler.__name__, self.objects[1].pk, 'my_field') in self.handler_backend.handlers)

        self.handler_backend.register(PostTransitionHandler, test_handler, self.objects[1], 'my_field')

        self.assertEqual(1, Handler.objects.filter(hash='%s.%s_object%sfield%s' % (PostTransitionHandler.__module__, PostTransitionHandler.__name__, self.objects[1].pk, 'my_field')).count())
        self.assertTrue('%s.%s_object%sfield%s' % (PostTransitionHandler.__module__, PostTransitionHandler.__name__, self.objects[1].pk, 'my_field') in self.handler_backend.handlers)
        self.assertEqual(test_handler.__name__, list(self.handler_backend.handlers.values())[0].__name__)

    def test_register_in_multiprocessing(self):
        from multiprocessing import Process, Queue

        self.handler_backend.register(PostTransitionHandler, test_handler, self.objects[1], 'my_field')
        self.assertEqual(1, Handler.objects.count())

        def worker2(q):
            second_handler_backend = load_handler_backend()
            handlers = second_handler_backend.get_handlers(PostTransitionHandler, self.objects[1], 'my_field')
            q.put([f.__name__ for f in handlers])

        q = Queue()
        p2 = Process(target=worker2, args=(q,))

        p2.start()

        handlers = q.get(timeout=1)
        self.assertEqual(1, len(handlers))
        self.assertEqual(test_handler.__name__, handlers[0])

    def test_get_handlers(self):
        self.handler_backend.register(PostTransitionHandler, test_handler, self.objects[1], 'my_field')
        handlers = self.handler_backend.get_handlers(PostTransitionHandler, self.objects[1], 'my_field')
        self.assertEqual(1, len(handlers))

        self.assertEqual(test_handler.__name__, handlers[0].__name__)

    def test_get_handlers_in_multiprocessing(self):
        from multiprocessing import Process, Queue

        Handler.objects.update_or_create(
            hash='%s.%s_object%sfield%s' % (PostTransitionHandler.__module__, PostTransitionHandler.__name__, self.objects[1].pk, 'my_field'),
            defaults={
                'method': '%s.%s' % (test_handler.__module__, test_handler.__name__),
                'handler_cls': '%s.%s' % (PostTransitionHandler.__module__, PostTransitionHandler.__name__),
            }
        )

        def worker2(q):
            handlers = self.handler_backend.get_handlers(PostTransitionHandler, self.objects[1], 'my_field')
            q.put([f.__name__ for f in handlers])

        q = Queue()
        p2 = Process(target=worker2, args=(q,))

        p2.start()

        handlers = q.get(timeout=1)
        self.assertEqual(1, len(handlers))
        self.assertEqual(test_handler.__name__, handlers[0])



__author__ = 'ahmetdal'

from .base_model import *
from .state import *
from .transition import *
from .proceeding_meta import *
from .proceeding import *
from .proceeding_track import *
from .handler import *



from django.db.models import Q

from river.models import FORWARD
from river.models.proceeding import Proceeding
from river.models.state import State
from river.utils.error_code import ErrorCode
from river.utils.exceptions import RiverException
from .proceeding import ProceedingService

__author__ = 'ahmetdal'


class StateService:
    def __init__(self):
        pass

    @staticmethod
    def get_state_by(**kwargs):
        if kwargs:
            return State.objects.get(**kwargs)

    @staticmethod
    def get_current_state(workflow_object, field):
        return getattr(workflow_object, field)

    @staticmethod
    def get_available_states(workflow_object, field, user, include_user=True):
        current_state = getattr(workflow_object, field)
        proceedings = Proceeding.objects.filter(
            field=field,
            workflow_object=workflow_object,
            meta__transition__source_state=current_state,
        )
        if include_user:
            proceedings = proceedings.filter(
                meta__permissions__in=user.user_permissions.all()
            )
        destination_states = proceedings.values_list('meta__transition__destination_state', flat=True)
        return State.objects.filter(pk__in=destination_states)

    @staticmethod
    def get_initial_state(content_type, field):
        """
        A state which is not a destination of a transition but can be source of a transition OR not (a destination of a transition and this transition direction is FORWARD)
        """
        states = State.objects.filter(pk__in=ProceedingService.get_initial_proceedings(content_type, field).values_list('meta__transition__source_state', flat=True))
        if states.count() == 0:
            raise RiverException(ErrorCode.NO_AVAILABLE_INITIAL_STATE, 'There is no available initial state for the content type %s. ' % content_type)
        elif states.count() > 1:
            raise RiverException(ErrorCode.MULTIPLE_INITIAL_STATE, 'There are multiple initial state for the content type %s. Have only one initial state' % content_type)

        return states[0]

    @staticmethod
    def get_final_states(content_type, field):

        """
        A state which is not a source of a transition but can be destination of a transition OR not (a source of a transition and this transition direction is FORWARD)
        """
        proceedings = ProceedingService.get_final_proceedings(content_type, field)
        if proceedings.count() == 0:
            raise RiverException(ErrorCode.NO_AVAILABLE_FINAL_STATE, 'There is no available final state for the content type %s.' % content_type)

        return State.objects.filter(pk__in=proceedings.values_list('meta__transition__destination_state', flat=True))



from django.test import TestCase

__author__ = 'ahmetdal'


class BaseTestCase(TestCase):
    def setUp(self):
        super(BaseTestCase, self).setUp()
        print('%s is initialized' % self.__class__)

    def tearDown(self):
        super(BaseTestCase, self).tearDown()
        print('%s is finished' % self.__class__)



from django.contrib.contenttypes.models import ContentType
from django.db.models.signals import m2m_changed

from river.models import Proceeding
from river.models.proceeding_meta import ProceedingMeta, post_group_change, post_permissions_change
from river.services.proceeding_meta import ProceedingMetaService
from river.tests.base_test import BaseTestCase
from river.tests.models import TestModel
from river.tests.models.factories import TestModelObjectFactory

__author__ = 'ahmetdal'


class test_ProceedingMetaService(BaseTestCase):
    def setUp(self):
        from river.models.factories import ProceedingMetaObjectFactory, StateObjectFactory

        self.state1 = StateObjectFactory()
        self.state2 = StateObjectFactory()
        self.state3 = StateObjectFactory()
        self.content_type = ContentType.objects.get_for_model(TestModel)

        self.proceeding_meta = ProceedingMetaObjectFactory(content_type=self.content_type, transition__source_state=self.state1, transition__destination_state=self.state2)
        self.object = TestModelObjectFactory().model
        self.field = "my_field"

    def test_apply_new_proceed_definition(self):
        from river.models.factories import ProceedingMetaObjectFactory, TransitionObjectFactory

        ct = self.proceeding_meta.content_type
        # self.assertEqual(0, Proceeding.objects.filter(workflow_object=self.object).count())
        # ObjectService.register_object(self.object, self.field)
        self.assertEqual(1, Proceeding.objects.filter(workflow_object=self.object).count())

        transition = TransitionObjectFactory(source_state=self.state2, destination_state=self.state3)

        m2m_changed.disconnect(post_group_change, ProceedingMeta.groups.through)
        m2m_changed.disconnect(post_permissions_change, ProceedingMeta.permissions.through)

        proceeding_meta = ProceedingMetaObjectFactory(content_type=ct, field=self.field, transition=transition, permissions__in=self.proceeding_meta.permissions.all())

        self.assertEqual(1, Proceeding.objects.filter(workflow_object=self.object, field=self.field).count())

        ProceedingMetaService.apply_new_proceeding_meta(proceeding_meta)

        self.assertEqual(2, Proceeding.objects.filter(workflow_object=self.object, field=self.field).count())

        proceeding_meta.save()

        self.assertEqual(2, Proceeding.objects.filter(workflow_object=self.object, field=self.field).count())



from django.conf import settings
from django.contrib.auth.models import Permission, Group
from django.contrib.auth.models import User
from django.contrib.contenttypes.models import ContentType

__author__ = 'ahmetdal'


class RiverConfig:
    # EMBEDDED = settings.get('RIVER_IS_EMBEDDED', True)
    CONTENT_TYPE_CLASS = getattr(settings, 'RIVER_CONTENT_TYPE_CLASS', ContentType)
    USER_CLASS = getattr(settings, 'RIVER_USER_CLASS', settings.AUTH_USER_MODEL)
    PERMISSION_CLASS = getattr(settings, 'RIVER_PERMISSION_CLASS', Permission)
    GROUP_CLASS = getattr(settings, 'RIVER_GROUP_CLASS', Group)
    HANDLER_BACKEND = getattr(settings, 'RIVER_HANDLER_BACKEND', {'backend': 'river.handlers.backends.memory.MemoryHandlerBackend'})
    HANDLER_BACKEND_CLASS = HANDLER_BACKEND.get('backend')
    HANDLER_BACKEND_CONFIG = HANDLER_BACKEND.get('config', {})



from django.contrib.contenttypes.models import ContentType
import factory

from river.tests.base_test import BaseTestCase
from river.tests.models import TestModel
from river.tests.models.factories import TestModelObjectFactory

__author__ = 'ahmetdal'


class ProceedingServiceBasedTest(BaseTestCase):
    def setUp(self):
        from river.models.factories import \
            TransitionObjectFactory, \
            UserObjectFactory, \
            PermissionObjectFactory, \
            ProceedingMetaObjectFactory, \
            StateObjectFactory

        TransitionObjectFactory.reset_sequence(0)
        ProceedingMetaObjectFactory.reset_sequence(0)
        StateObjectFactory.reset_sequence(0)
        TestModel.objects.all().delete()

        self.content_type = ContentType.objects.get_for_model(TestModel)
        self.permissions = PermissionObjectFactory.create_batch(4)
        self.user1 = UserObjectFactory(user_permissions=[self.permissions[0]])
        self.user2 = UserObjectFactory(user_permissions=[self.permissions[1]])
        self.user3 = UserObjectFactory(user_permissions=[self.permissions[2]])
        self.user4 = UserObjectFactory(user_permissions=[self.permissions[3]])

        self.field = 'my_field'
        self.states = StateObjectFactory.create_batch(
            9,
            label=factory.Sequence(lambda n: "s%s" % str(n + 1) if n <= 4 else ("s4.%s" % str(n - 4) if n <= 6 else "s5.%s" % str(n - 6)))
        )
        self.transitions = TransitionObjectFactory.create_batch(8,
                                                                source_state=factory.Sequence(
                                                                    lambda n: self.states[n] if n <= 2 else (self.states[n - 1]) if n <= 4 else (self.states[n - 2] if n <= 6 else self.states[4])),
                                                                destination_state=factory.Sequence(lambda n: self.states[n + 1]))

        self.proceeding_metas = ProceedingMetaObjectFactory.create_batch(
            9,
            content_type=self.content_type,
            field=self.field,
            transition=factory.Sequence(lambda n: self.transitions[n] if n <= 1 else self.transitions[n - 1]),
            order=factory.Sequence(lambda n: 1 if n == 2 else 0)
        )

        for n, proceeding_meta in enumerate(self.proceeding_metas):
            proceeding_meta.permissions.add(self.permissions[n] if n <= 3 else self.permissions[3])

        self.objects = TestModelObjectFactory.create_batch(2)



from django.dispatch import Signal
import logging

__author__ = 'ahmetdal'

pre_final = Signal(providing_args=["workflow_object", "field"])
post_final = Signal(providing_args=["workflow_object", "field"])

pre_transition = Signal(providing_args=["workflow_object", "field", "source_state", "destination_state", "proceeding"])
post_transition = Signal(providing_args=["workflow_object", "field", "source_state", "destination_state", "proceeding"])

pre_proceed = Signal(providing_args=["workflow_object", "field", "proceeding", "track"])
post_proceed = Signal(providing_args=["workflow_object", "field", "proceeding", "track"])

LOGGER = logging.getLogger(__name__)


class TransitionSignal(object):
    def __init__(self, status, workflow_object, field, proceeding):
        self.status = status
        self.workflow_object = workflow_object
        self.field = field
        self.proceeding = proceeding

    def __enter__(self):
        if self.status:
            pre_transition.send(
                sender=TransitionSignal.__class__,
                workflow_object=self.workflow_object,
                field=self.field,
                source_state=self.proceeding.meta.transition.source_state,
                destination_state=self.proceeding.meta.transition.destination_state,
                proceeding=self.proceeding,
            )
            LOGGER.debug("Pre transition signal IS sent for workflow object %s for field %s for transition %s -> %s" % (
                self.workflow_object, self.field, self.proceeding.meta.transition.source_state.label, self.proceeding.meta.transition.destination_state.label))
        else:
            LOGGER.debug("Pre transition signal IS NOT sent for workflow object %s for field %s. Although proceeding is occurred, not transition is occurred." % (self.workflow_object, self.field))

    def __exit__(self, type, value, traceback):
        if self.status:
            post_transition.send(
                sender=TransitionSignal.__class__,
                workflow_object=self.workflow_object,
                field=self.field,
                source_state=self.proceeding.meta.transition.source_state,
                destination_state=self.proceeding.meta.transition.destination_state,
                proceeding=self.proceeding,
            )
            LOGGER.debug("Post transition signal IS sent for workflow object %s for field %s for transition %s -> %s" % (
                self.workflow_object, self.field, self.proceeding.meta.transition.source_state.label, self.proceeding.meta.transition.destination_state.label))
        else:
            LOGGER.debug("Post transition signal IS NOT sent for workflow object %s for field %s. Although proceeding is occurred, not transition is occurred." % (self.workflow_object, self.field))


class ProceedingSignal(object):
    def __init__(self, workflow_object, field, proceeding, track):
        self.workflow_object = workflow_object
        self.field = field
        self.proceeding = proceeding
        self.track = track

    def __enter__(self):
        pre_proceed.send(
            sender=ProceedingSignal.__class__,
            workflow_object=self.workflow_object,
            field=self.field,
            proceeding=self.proceeding,
            track=self.track
        )
        LOGGER.debug("Pre proceeding signal IS sent for workflow object %s for field %s for transition %s -> %s" % (
            self.workflow_object, self.field, self.proceeding.meta.transition.source_state.label, self.proceeding.meta.transition.destination_state.label))

    def __exit__(self, type, value, traceback):
        post_proceed.send(
            sender=TransitionSignal.__class__,
            workflow_object=self.workflow_object,
            field=self.field,
            proceeding=self.proceeding,
            track=self.track
        )
        LOGGER.debug("Post proceeding signal IS sent for workflow object %s for field %s for transition %s -> %s" % (
            self.workflow_object, self.field, self.proceeding.meta.transition.source_state.label, self.proceeding.meta.transition.destination_state.label))


class FinalSignal(object):
    def __init__(self, workflow_object, field):
        self.workflow_object = workflow_object
        self.field = field
        self.status = self.workflow_object.on_final_state

    def __enter__(self):
        if self.status:
            pre_final.send(
                sender=FinalSignal.__class__,
                workflow_object=self.workflow_object,
                field=self.field
            )
            LOGGER.debug("Pre final signal IS sent for workflow object %s for field %s for final state %s" % (self.workflow_object, self.field, getattr(self.workflow_object, self.field).label))

    def __exit__(self, type, value, traceback):
        if self.status:
            post_final.send(
                sender=TransitionSignal.__class__,
                workflow_object=self.workflow_object,
                field=self.field,
            )
            LOGGER.debug("Post final signal IS sent for workflow object %s for field %s for final state %s" % (self.workflow_object, self.field, getattr(self.workflow_object, self.field).label))



from datetime import datetime
import logging

from django.db.transaction import atomic

from river.models.proceeding import APPROVED, PENDING
from river.services.proceeding import ProceedingService
from river.services.state import StateService
from river.signals import ProceedingSignal, TransitionSignal, FinalSignal
from river.utils.error_code import ErrorCode
from river.utils.exceptions import RiverException

__author__ = 'ahmetdal'


class TTSSignal(object):
    def __init__(self, status, pre_signal, post_signal, **kwargs):
        self.status = status
        self.pre_signal = pre_signal
        self.post_signal = post_signal
        self.kwargs = kwargs

    def __enter__(self):
        if self.status:
            self.pre_signal.send(
                sender=TTSSignal.__class__,
                **self.kwargs
            )

    def __exit__(self, type, value, traceback):
        if self.status:
            self.post_signal.send(
                sender=TTSSignal.__class__,
                **self.kwargs
            )


LOGGER = logging.getLogger(__name__)


class TransitionService(object):
    def __init__(self):
        pass

    @staticmethod
    @atomic
    def proceed(workflow_object, field, user, next_state=None, god_mod=False):

        def process(workflow_object, field, user, action, next_state=None, god_mod=False):
            current_state = getattr(workflow_object, field)
            proceedings = ProceedingService.get_available_proceedings(workflow_object, field, [current_state], user=user, god_mod=god_mod)
            c = proceedings.count()
            if c == 0:
                raise RiverException(ErrorCode.NO_AVAILABLE_NEXT_STATE_FOR_USER, "There is no available state for destination for the user.")
            if c > 1:
                if next_state:
                    proceedings = proceedings.filter(meta__transition__destination_state=next_state)
                    if proceedings.count() == 0:
                        available_states = StateService.get_available_states(workflow_object, field, user)
                        raise RiverException(ErrorCode.INVALID_NEXT_STATE_FOR_USER,
                                             "Invalid state is given(%s). Valid states is(are) %s" % (next_state.__unicode__(), ','.join([ast.__unicode__() for ast in available_states])))
                else:
                    raise RiverException(ErrorCode.NEXT_STATE_IS_REQUIRED, "State must be given when there are multiple states for destination")
            proceeding = proceedings[0]
            proceeding.status = action
            proceeding.transactioner = user
            proceeding.transaction_date = datetime.now()
            proceeding.save()

            c = False
            track = workflow_object.proceeding_track
            while not c:
                track, c = proceeding.tracks.get_or_create(previous_track=track)
            return proceeding, track

        proceeding, track = process(workflow_object, field, user, APPROVED, next_state, god_mod)

        current_state = getattr(workflow_object, field)
        # Any other proceeding is left?
        required_proceedings = ProceedingService.get_available_proceedings(workflow_object, field, [current_state], destination_state=next_state, god_mod=god_mod)

        transition_status = False
        if required_proceedings.count() == 0:
            setattr(workflow_object, field, proceeding.meta.transition.destination_state)
            transition_status = True

            # Next states should be PENDING back again if there is circle.
            ProceedingService.get_next_proceedings(workflow_object, field).update(status=PENDING)

        with ProceedingSignal(workflow_object, field, proceeding, track), TransitionSignal(transition_status, workflow_object, field, proceeding), FinalSignal(workflow_object, field):
            workflow_object.save()

        LOGGER.debug("Workflow object %s for field %s is proceeded for next transition. Transition: %s -> %s" % (workflow_object, field, current_state.label, getattr(workflow_object, field).label))



from river.models import State, Proceeding
from river.services.proceeding import ProceedingService
from river.services.object import ObjectService
from river.tests.services.proceeding_service_based_test import ProceedingServiceBasedTest

__author__ = 'ahmetdal'


class test_ProceedingService(ProceedingServiceBasedTest):
    def test_get_proceedings_object_waiting_for_approval_without_skip(self):
        ObjectService.register_object(self.objects[0], self.field)
        ObjectService.register_object(self.objects[1], self.field)

        proceedings = ProceedingService.get_available_proceedings(self.objects[0], self.field, [self.objects[0].my_field], user=self.user1)
        self.assertEqual(1, proceedings.count())

        proceedings = ProceedingService.get_available_proceedings(self.objects[0], self.field, [self.objects[0].my_field], user=self.user2)
        self.assertEqual(0, proceedings.count())

        proceedings = ProceedingService.get_available_proceedings(self.objects[0], self.field, [self.objects[0].my_field], user=self.user3)
        self.assertEqual(0, proceedings.count())

        proceedings = ProceedingService.get_available_proceedings(self.objects[0], self.field, [self.objects[0].my_field], user=self.user4)
        self.assertEqual(0, proceedings.count())

    def test_get_next_proceedings(self):
        ObjectService.register_object(self.objects[0], self.field)
        ObjectService.register_object(self.objects[1], self.field)

        proceedings = ProceedingService.get_next_proceedings(self.objects[0], self.field)
        self.assertEqual(9, proceedings.count())

        self.objects[0].proceed(self.user1)

        proceedings = ProceedingService.get_next_proceedings(self.objects[0], self.field)
        self.assertEqual(8, proceedings.count())

        self.objects[0].proceed(self.user2)


        # Two proceedings exist on same level
        proceedings = ProceedingService.get_next_proceedings(self.objects[0], self.field)
        self.assertEqual(8, proceedings.count())

        self.objects[0].proceed(self.user3)

        proceedings = ProceedingService.get_next_proceedings(self.objects[0], self.field)
        self.assertEqual(6, proceedings.count())

        self.objects[0].proceed(self.user4, next_state=State.objects.get(label='s4'))
        proceedings = ProceedingService.get_next_proceedings(self.objects[0], self.field)
        self.assertEqual(2, proceedings.count())

        self.objects[0].proceed(self.user4, next_state=State.objects.get(label='s4.1'))
        proceedings = ProceedingService.get_next_proceedings(self.objects[0], self.field)
        self.assertEqual(0, proceedings.count())

    def test_get_proceedings_object_waiting_for_approval_with_skip(self):
        ObjectService.register_object(self.objects[0], self.field)
        ObjectService.register_object(self.objects[1], self.field)

        proceedings = self.objects[1].get_available_proceedings(self.user1)
        self.assertEqual(1, proceedings.count())
        self.assertEqual(State.objects.get(label='s2'), proceedings[0].meta.transition.destination_state)

        Proceeding.objects.filter(
            field=self.field,
            workflow_object=self.objects[1],
            meta__transition__destination_state=State.objects.get(label='s2')
        ).update(skip=True)

        proceedings = ProceedingService.get_available_proceedings(self.objects[1], self.field, [self.objects[1].my_field], user=self.user2)
        self.assertEqual(1, proceedings.count())
        self.assertEqual(State.objects.get(label='s3'), proceedings[0].meta.transition.destination_state)

        Proceeding.objects.filter(
            field=self.field,
            workflow_object=self.objects[1],
            meta__transition__destination_state=State.objects.get(label='s3')
        ).update(skip=True)

        proceedings = ProceedingService.get_available_proceedings(self.objects[1], self.field, [self.objects[1].my_field], user=self.user4)
        self.assertEqual(2, proceedings.count())
        self.assertEqual(State.objects.get(label='s4'), proceedings[0].meta.transition.destination_state)
        self.assertEqual(State.objects.get(label='s5'), proceedings[1].meta.transition.destination_state)

        Proceeding.objects.filter(
            field=self.field,
            workflow_object=self.objects[1],
            meta__transition__destination_state=State.objects.get(label='s4')
        ).update(skip=True)

        proceedings = ProceedingService.get_available_proceedings(self.objects[1], self.field, [self.objects[1].my_field], user=self.user4)
        self.assertEqual(3, proceedings.count())
        self.assertEqual(State.objects.get(label='s5'), proceedings[0].meta.transition.destination_state)
        self.assertEqual(State.objects.get(label='s4.1'), proceedings[1].meta.transition.destination_state)
        self.assertEqual(State.objects.get(label='s4.2'), proceedings[2].meta.transition.destination_state)

        Proceeding.objects.filter(
            field=self.field,
            workflow_object=self.objects[1],
            meta__transition__destination_state=State.objects.get(label='s4')
        ).update(skip=False)
        Proceeding.objects.filter(
            field=self.field,
            workflow_object=self.objects[1],
            meta__transition__destination_state=State.objects.get(label='s5')
        ).update(skip=True)

        proceedings = ProceedingService.get_available_proceedings(self.objects[1], self.field, [self.objects[1].my_field], user=self.user4)
        self.assertEqual(3, proceedings.count())
        self.assertEqual(State.objects.get(label='s4'), proceedings[0].meta.transition.destination_state)
        self.assertEqual(State.objects.get(label='s5.1'), proceedings[1].meta.transition.destination_state)
        self.assertEqual(State.objects.get(label='s5.2'), proceedings[2].meta.transition.destination_state)

        Proceeding.objects.filter(
            field=self.field,
            workflow_object=self.objects[1],
            meta__transition__destination_state__in=State.objects.filter(label__in=['s4', 's5'])
        ).update(skip=True)

        proceedings = ProceedingService.get_available_proceedings(self.objects[1], self.field, [self.objects[1].my_field], user=self.user4)
        self.assertEqual(4, proceedings.count())
        self.assertEqual(State.objects.get(label='s4.1'), proceedings[0].meta.transition.destination_state)
        self.assertEqual(State.objects.get(label='s4.2'), proceedings[1].meta.transition.destination_state)
        self.assertEqual(State.objects.get(label='s5.1'), proceedings[2].meta.transition.destination_state)
        self.assertEqual(State.objects.get(label='s5.2'), proceedings[3].meta.transition.destination_state)

        Proceeding.objects.filter(
            field=self.field,
            workflow_object=self.objects[1],
            meta__transition__destination_state__in=State.objects.filter(label__in=['s4.1', 's5.1'])
        ).update(skip=True)

        proceedings = ProceedingService.get_available_proceedings(self.objects[1], self.field, [self.objects[1].my_field], user=self.user4)
        self.assertEqual(2, proceedings.count())
        self.assertEqual(State.objects.get(label='s4.2'), proceedings[0].meta.transition.destination_state)
        self.assertEqual(State.objects.get(label='s5.2'), proceedings[1].meta.transition.destination_state)



from django.db import models
from django.utils.translation import ugettext_lazy as _
from mptt.fields import TreeOneToOneField

from river.models import Proceeding
from river.models.base_model import BaseModel

__author__ = 'ahmetdal'


class ProceedingTrack(BaseModel):
    class Meta:
        app_label = 'river'
        verbose_name = _("Proceeding Track")
        verbose_name_plural = _("Proceeding Tracks")

    proceeding = models.ForeignKey(Proceeding, verbose_name=_('Proceeding'), related_name="tracks")

    process_date = models.DateTimeField(auto_now_add=True, null=True, blank=True)

    previous_track = TreeOneToOneField("self", verbose_name=_('Previous track'), related_name="next_track", null=True, blank=True)





from river.models import ProceedingMeta
from river.models.proceeding import Proceeding

__author__ = 'ahmetdal'


class ProceedingMetaService(object):
    @staticmethod
    def apply_new_proceeding_meta(new_proceeding_meta):
        if new_proceeding_meta.proceedings.count() == 0:
            content_type = new_proceeding_meta.content_type
            WorkflowObjectClass = content_type.model_class()
            field = new_proceeding_meta.field
            Proceeding.objects.bulk_create(
                list(
                    Proceeding(
                        workflow_object=workflow_object,
                        meta=new_proceeding_meta,
                        field=field,
                        content_type=content_type
                    )
                    for workflow_object in WorkflowObjectClass.objects.all()
                )
            )

    @staticmethod
    def build_tree(instance):
        parents = ProceedingMeta.objects.filter(transition__destination_state__pk=instance.transition.source_state.pk).exclude(pk__in=instance.parents.values_list('pk', flat=True))
        children = ProceedingMeta.objects.filter(transition__source_state__pk=instance.transition.destination_state.pk).exclude(parents__in=[instance.pk])
        instance = ProceedingMeta.objects.get(pk=instance.pk)
        if parents:
            instance.parents.add(*parents)

        for child in children:
            child.parents.add(instance)



from django.db import models
from django.utils.translation import ugettext_lazy as _

from river.models.base_model import BaseModel
from river.models.state import State
from river.services.config import RiverConfig

__author__ = 'ahmetdal'


class TransitionManager(models.Manager):
    def get_by_natural_key(self, source_state, destination_state):
        return self.get(source_state=State.objects.get_by_natural_key(source_state), destination_state=State.objects.get_by_natural_key(destination_state))


BACKWARD = 0
FORWARD = 1

DIRECTIONS = [
    (BACKWARD, _("Backward")),
    (FORWARD, _("Forward"))
]


class Transition(BaseModel):
    class Meta:
        app_label = 'river'
        verbose_name = _("Transition")
        verbose_name_plural = _("Transitions")
        unique_together = (('source_state', 'destination_state'),)

    objects = TransitionManager()

    source_state = models.ForeignKey(State, verbose_name=_("Source State"), related_name='transitions_as_source')
    destination_state = models.ForeignKey(State, verbose_name=_("Next State"), related_name='transitions_as_destination')
    direction = models.SmallIntegerField(_("Transition Direction"), default=FORWARD)

    def natural_key(self):
        return self.source_state.slug, self.destination_state.slug

    def __unicode__(self):
        return '%s -> %s' % (self.source_state, self.destination_state)



from river.handlers.handler import Handler
from river.signals import pre_transition, post_transition

__author__ = 'ahmetdal'


class TransitionHandler(Handler):
    @classmethod
    def get_hash(cls, workflow_object, field, source_state=None, destination_state=None, proceeding=None, *args, **kwargs):
        return super(TransitionHandler, cls).get_hash(workflow_object, field) + ('source_state' + str(source_state.pk) if source_state else '') + (
            'destination_state' + str(destination_state.pk) if destination_state else '') + ('proceeding' + str(proceeding.pk) if proceeding else '')


class PreTransitionHandler(TransitionHandler):
    pass


class PostTransitionHandler(TransitionHandler):
    pass


pre_transition.connect(PreTransitionHandler.dispatch)
post_transition.connect(PostTransitionHandler.dispatch)



from river.models.proceeding import Proceeding
from river.services.proceeding import ProceedingService

__author__ = 'ahmetdal'


# noinspection PyClassHasNoInit
class ObjectService:
    @staticmethod
    def register_object(workflow_object, field):
        proceedings = Proceeding.objects.filter(workflow_object=workflow_object, field=field)
        if proceedings.count() == 0:
            ProceedingService.init_proceedings(workflow_object, field)

    @staticmethod
    def get_objects_waiting_for_approval(content_type, field, user):
        object_pks = []
        WorkflowObjectClass = content_type.model_class()
        for workflow_object in WorkflowObjectClass.objects.all():
            current_state = getattr(workflow_object, field)
            proceedings = ProceedingService.get_available_proceedings(workflow_object, field, [current_state], user=user)
            if proceedings.count():
                object_pks.append(workflow_object.pk)
        return WorkflowObjectClass.objects.filter(pk__in=object_pks)

    @staticmethod
    def get_object_count_waiting_for_approval(content_type, field, user):
        return ObjectService.get_objects_waiting_for_approval(content_type, field, user).count()

    @staticmethod
    def is_workflow_completed(workflow_object, field):
        current_state = getattr(workflow_object, field)
        return Proceeding.objects.filter(workflow_object=workflow_object, meta__transition__source_state=current_state).count() == 0



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from river.models.factories import ProceedingMetaObjectFactory
from river.services.proceeding_meta import ProceedingMetaService


def build_tree(apps, schema_editor):
    # We can't import the Person model directly as it may be a newer
    # version than this migration expects. We use the historical version.
    ProceedingMeta = apps.get_model("river", "ProceedingMeta")

    for pm in ProceedingMeta.objects.all():
        ProceedingMetaService.build_tree(pm)


class Migration(migrations.Migration):
    dependencies = [
        ('river', '0002_auto_20150916_0144'),
    ]

    operations = [
        migrations.AddField(
            model_name='proceedingmeta',
            name='parents',
            field=models.ManyToManyField(related_name='children', verbose_name=b'parents', to='river.ProceedingMeta', db_index=True),
        ),
        migrations.RunPython(build_tree)
    ]



from river.services.config import RiverConfig

__author__ = 'ahmetdal'


def load_handler_backend():
    handler_module, handler_cls = RiverConfig.HANDLER_BACKEND_CLASS.rsplit('.', 1)
    return getattr(__import__(handler_module, fromlist=[handler_cls]), handler_cls)(**RiverConfig.HANDLER_BACKEND_CONFIG)


handler_backend = load_handler_backend()



try:
    from django.contrib.contenttypes.fields import GenericForeignKey
except ImportError:
    from django.contrib.contenttypes.generic import GenericForeignKey

from django.db import models

from django.utils.translation import ugettext_lazy as _

from river.models.proceeding_meta import ProceedingMeta
from river.models.base_model import BaseModel
from river.models.managers.proceeding import ProceedingManager
from river.services.config import RiverConfig

__author__ = 'ahmetdal'

PENDING = 0
APPROVED = 1
REJECTED = 2

PROCEEDING_STATUSES = [
    (PENDING, _('Pending')),
    (APPROVED, _('Approved')),
    (REJECTED, _('Rejected')),
]


class Proceeding(BaseModel):
    class Meta:
        app_label = 'river'
        verbose_name = _("Proceeding")
        verbose_name_plural = _("Proceedings")

    objects = ProceedingManager()

    content_type = models.ForeignKey(RiverConfig.CONTENT_TYPE_CLASS, verbose_name=_('Content Type'))
    object_id = models.PositiveIntegerField(verbose_name=_('Related Object'))
    field = models.CharField(verbose_name=_('Field'), max_length=200)
    workflow_object = GenericForeignKey('content_type', 'object_id')

    meta = models.ForeignKey(ProceedingMeta, verbose_name=_('Meta'), related_name="proceedings")
    transactioner = models.ForeignKey(RiverConfig.USER_CLASS, verbose_name=_('Transactioner'), null=True, blank=True)
    transaction_date = models.DateTimeField(null=True, blank=True)

    status = models.IntegerField(_('Status'), choices=PROCEEDING_STATUSES, default=PENDING)

    skip = models.BooleanField(_('Skip'), default=False)

    permissions = models.ManyToManyField(RiverConfig.PERMISSION_CLASS, verbose_name=_('Permissions'))
    groups = models.ManyToManyField(RiverConfig.GROUP_CLASS, verbose_name=_('Groups'))
    order = models.IntegerField(default=0, verbose_name=_('Order'))

    enabled = models.BooleanField(_('Enabled?'), default=True)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import uuid

from django.db import models, migrations
from django.utils.text import slugify


def move_from_transition_to_meta(apps, schema_editor):
    # We can't import the Person model directly as it may be a newer
    # version than this migration expects. We use the historical version.
    ProceedingMeta = apps.get_model("river", "ProceedingMeta")
    for pm in ProceedingMeta.objects.all():
        pm.content_type = pm.transition.content_type
        pm.field = pm.transition.field
        pm.save()


def set_slug(apps, schema_editor):
    # We can't import the Person model directly as it may be a newer
    # version than this migration expects. We use the historical version.
    State = apps.get_model("river", "State")

    for s in State.objects.all():
        s.slug = slugify(s.label)
        s.save()


class Migration(migrations.Migration):
    dependencies = [
        ('river', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='proceedingmeta',
            name='content_type',
            field=models.ForeignKey(default=-1, verbose_name='Content Type', to='contenttypes.ContentType'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='proceedingmeta',
            name='field',
            field=models.CharField(default='state', max_length=200, verbose_name='Field'),
            preserve_default=False,
        ),
        migrations.RunPython(move_from_transition_to_meta),

        migrations.RemoveField(
            model_name='transition',
            name='content_type',
        ),
        migrations.RemoveField(
            model_name='transition',
            name='field',
        ),
        migrations.AlterUniqueTogether(
            name='proceedingmeta',
            unique_together=set([('content_type', 'field', 'transition', 'order')]),
        ),
        migrations.AlterUniqueTogether(
            name='transition',
            unique_together=set([('source_state', 'destination_state')]),
        ),
        migrations.AddField(
            model_name='state',
            name='slug',
            field=models.SlugField(null=True, blank=True, unique=True),
            preserve_default=True
        ),
        migrations.RunPython(set_slug),
        migrations.AlterField(
            model_name='handler',
            name='date_created',
            field=models.DateTimeField(auto_now_add=True, verbose_name='Date Created', null=True),
        ),
        migrations.AlterField(
            model_name='handler',
            name='date_updated',
            field=models.DateTimeField(auto_now=True, verbose_name='Date Updated', null=True),
        ),
        migrations.AlterField(
            model_name='proceeding',
            name='date_created',
            field=models.DateTimeField(auto_now_add=True, verbose_name='Date Created', null=True),
        ),
        migrations.AlterField(
            model_name='proceeding',
            name='date_updated',
            field=models.DateTimeField(auto_now=True, verbose_name='Date Updated', null=True),
        ),
        migrations.AlterField(
            model_name='proceedingmeta',
            name='date_created',
            field=models.DateTimeField(auto_now_add=True, verbose_name='Date Created', null=True),
        ),
        migrations.AlterField(
            model_name='proceedingmeta',
            name='date_updated',
            field=models.DateTimeField(auto_now=True, verbose_name='Date Updated', null=True),
        ),
        migrations.AlterField(
            model_name='proceedingtrack',
            name='date_created',
            field=models.DateTimeField(auto_now_add=True, verbose_name='Date Created', null=True),
        ),
        migrations.AlterField(
            model_name='proceedingtrack',
            name='date_updated',
            field=models.DateTimeField(auto_now=True, verbose_name='Date Updated', null=True),
        ),
        migrations.AlterField(
            model_name='state',
            name='date_created',
            field=models.DateTimeField(auto_now_add=True, verbose_name='Date Created', null=True),
        ),
        migrations.AlterField(
            model_name='state',
            name='date_updated',
            field=models.DateTimeField(auto_now=True, verbose_name='Date Updated', null=True),
        ),
        migrations.AlterField(
            model_name='transition',
            name='date_created',
            field=models.DateTimeField(auto_now_add=True, verbose_name='Date Created', null=True),
        ),
        migrations.AlterField(
            model_name='transition',
            name='date_updated',
            field=models.DateTimeField(auto_now=True, verbose_name='Date Updated', null=True),
        ),
    ]



__author__ = 'ahmetdal'


class RiverException(Exception):
    code = None

    def __init__(self, error_code, *args, **kwargs):
        super(RiverException, self).__init__(*args, **kwargs)
        self.code = error_code



import logging

from django.db.models import Min, Q
from django.contrib import auth

from river.models.proceeding import Proceeding, PENDING
from river.models.proceeding_meta import ProceedingMeta
from river.models.state import State
from river.services.config import RiverConfig

__author__ = 'ahmetdal'

LOGGER = logging.getLogger(__name__)


class ProceedingService(object):
    @staticmethod
    def init_proceedings(workflow_object, field):

        content_type = RiverConfig.CONTENT_TYPE_CLASS.objects.get_for_model(workflow_object)
        for proceeding_meta in ProceedingMeta.objects.filter(content_type=content_type, field=field):
            proceeding, created = Proceeding.objects.update_or_create(
                meta=proceeding_meta,
                field=proceeding_meta.field,
                workflow_object=workflow_object,
                defaults={
                    'order': proceeding_meta.order,
                    'status': PENDING,
                }
            )
            proceeding.permissions.add(*proceeding_meta.permissions.all())
            proceeding.groups.add(*proceeding_meta.groups.all())

        workflow_object.save()
        LOGGER.debug("Proceedings are initialized for workflow object %s and field %s" % (workflow_object, field))

    @staticmethod
    def get_available_proceedings(workflow_object, field, source_states, user=None, destination_state=None, god_mod=False):

        def get_proceeding(proceedings):
            min_order = proceedings.aggregate(Min('order'))['order__min']
            proceedings = proceedings.filter(order=min_order)

            if destination_state:
                proceedings = proceedings.filter(meta__transition__destination_state=destination_state)

            return proceedings

        def authorize_proceedings(proceedings):
            group_q = Q()
            for g in user.groups.all():
                group_q = group_q | Q(groups__in=[g])

            permissions = []
            for backend in auth.get_backends():
                permissions.extend(backend.get_all_permissions(user))

            permission_q = Q()
            for p in permissions:
                label, codename = p.split('.')
                permission_q = permission_q | Q(permissions__content_type__app_label=label, permissions__codename=codename)

            return proceedings.filter(
                (
                    (Q(transactioner__isnull=True) | Q(transactioner=user)) &
                    (Q(permissions__isnull=True) | permission_q) &
                    (Q(groups__isnull=True) | group_q)
                )
            )

        proceedings = Proceeding.objects.filter(
            workflow_object=workflow_object,
            field=field,
            meta__transition__source_state__in=source_states,
            status=PENDING,
            enabled=True
        )

        suitable_proceedings = get_proceeding(proceedings.filter(skip=False))

        if user and not god_mod:
            suitable_proceedings = authorize_proceedings(suitable_proceedings)

        skipped_proceedings = get_proceeding(proceedings.filter(skip=True))
        if skipped_proceedings:
            source_state_pks = list(skipped_proceedings.values_list('meta__transition__destination_state', flat=True))
            suitable_proceedings = suitable_proceedings | ProceedingService.get_available_proceedings(workflow_object, field, State.objects.filter(pk__in=source_state_pks),
                                                                                                      user=user, destination_state=destination_state, god_mod=god_mod)
        return suitable_proceedings

    @staticmethod
    def get_next_proceedings(workflow_object, field, proceeding_pks=None, current_states=None, index=0, limit=None):
        if not proceeding_pks:
            proceeding_pks = []
        index += 1
        current_states = list(current_states.values_list('pk', flat=True)) if current_states else [getattr(workflow_object, field)]
        next_proceedings = Proceeding.objects.filter(workflow_object=workflow_object, field=field, meta__transition__source_state__in=current_states)
        if next_proceedings.exists() and not next_proceedings.filter(pk__in=proceeding_pks).exists() and (not limit or index < limit):
            proceedings = ProceedingService.get_next_proceedings(
                workflow_object,
                field,
                proceeding_pks=proceeding_pks + list(next_proceedings.values_list('pk', flat=True)),
                current_states=State.objects.filter(pk__in=next_proceedings.values_list('meta__transition__destination_state', flat=True)),
                index=index,
                limit=limit
            )
        else:
            proceedings = Proceeding.objects.filter(pk__in=proceeding_pks)

        return proceedings

    @staticmethod
    def has_user_any_action(content_type, field, user):
        """
        :param content_type_id:
        :param field_id:
        :param user_id:
        :return: Boolean value indicates whether the user has any role for the content type and field are sent. Any elements existence
          accepted, rejected or pending for the user, means the user in active for the content type and field.
        """
        proceedings = Proceeding.objects.filter(Q(transactioner=user) | Q(permissions__in=user.user_permissions.all()) | Q(groups__in=user.groups.all())).filter(content_type=content_type,
                                                                                                                                                                 field=field)
        return proceedings.count() != 0

    @staticmethod
    def override_permissions(proceeding, permissions):
        proceeding.permissions.clear()
        proceeding.permissions.add(*permissions)

    @staticmethod
    def override_groups(proceeding, groups):
        proceeding.groups.clear()
        proceeding.groups.add(*groups)

    @staticmethod
    def get_initial_proceedings(content_type, field):
        return Proceeding.objects.filter(meta__parents__isnull=True)

    @staticmethod
    def get_final_proceedings(content_type, field):
        return Proceeding.objects.filter(meta__children__isnull=True)



from django.db import models
from river.models.fields.state import StateField

__author__ = 'ahmetdal'


class TestModel(models.Model):
    test_field = models.CharField(max_length=50, null=True, blank=True)
    my_field = StateField()



from river.models import Proceeding
from river.services.object import ObjectService
from river.services.state import StateService
from river.tests.services.proceeding_service_based_test import ProceedingServiceBasedTest

__author__ = 'ahmetdal'


# noinspection PyPep8Naming
class test_ObjectService(ProceedingServiceBasedTest):
    def test_init(self):
        ObjectService.register_object(self.objects[0], self.field)
        ObjectService.register_object(self.objects[1], self.field)

        initial_state = StateService.get_initial_state(self.content_type, self.field)

        self.assertEqual(initial_state, getattr(self.objects[0], self.field))
        self.assertEqual(18, Proceeding.objects.count())

    def test_get_objects_waiting_for_approval_for_user(self):
        on_approval_objects = ObjectService.get_objects_waiting_for_approval(self.content_type, self.field, self.user1)
        self.assertEqual(2, on_approval_objects.count())
        self.assertEqual(self.objects[0], on_approval_objects[0])

        on_approval_objects = ObjectService.get_objects_waiting_for_approval(self.content_type, self.field, self.user2)
        self.assertEqual(0, on_approval_objects.count())

        on_approval_objects = ObjectService.get_objects_waiting_for_approval(self.content_type, self.field, self.user3)
        self.assertEqual(0, on_approval_objects.count())

        on_approval_objects = ObjectService.get_objects_waiting_for_approval(self.content_type, self.field, self.user4)
        self.assertEqual(0, on_approval_objects.count())
